# src/koyaneframework/banner.py

import pyfiglet

def show_banner():
    ascii_banner = pyfiglet.figlet_format("KYF", font="slant")
    print(ascii_banner)
    print("Koyane-Framework :: wordlist forge & analysis toolkit")
    print ("made by Puppetm4ster")
    print()
    print()



from  pathlib import Path
from koyaneframework.utils import external_sort, SPECIAL_CHARACTERS
class ContentAnalyzer:

    file_path: Path = None
    wordlines = 0

    smallestword = None
    biggestword = None

    av_wordlen = 0

    char_count = {}

    dublicate_words = []

    DIGIT = "Has digit"
    UPPER = "Has upper"
    SPECSIGN = "Has special character"

    DIGITANDUPPER = "Has digit & upper"
    DIGITANDSPECSIGN = "Has digit & special character"
    UPPERANDSPECSIGN = "Has upper & special character"

    DIGITUPPERSPECSIGN = "Has digit & upper & special character"

    pw_type_list = {
        DIGIT: 0,
        UPPER: 0,
        SPECSIGN: 0,

        DIGITANDUPPER: 0,
        DIGITANDSPECSIGN: 0,
        UPPERANDSPECSIGN: 0,

        DIGITUPPERSPECSIGN: 0

    }

    def __init__(self, file_path: Path):
        self.file_path = file_path



    def analyze_file(self, count: bool=False, min_max:bool=False, av_length: bool=False, char_freq: bool=False, av_entropy: bool=False,dupl: bool=False, weak_pw: bool=False, perc_pw: bool=False):
        output_file = self.file_path
        output_file = Path(__file__).resolve().parents[1] / "tmp" / "sorted_list.kyftmp"
        if dupl:
            external_sort(self.file_path, output_file)
        
        with open(output_file, "r", encoding="utf-8") as file:
            # helper var for calculating $av_wordlen
            totalwordlen = 0
            last_word = None
            for line in file:
                word = line
                # count flag
                if count or av_length:
                    self.wordlines += 1
                if min_max:
                    self._password_min_max_info(word)
                if av_length:
                    totalwordlen += len(word)
                if char_freq:
                    self._char_frequency(word)
                if av_entropy:
                    # displays average password-entropy
                    pass
                if dupl:
                    self._password_duplicate(last_word,word)
                    last_word = word
                if weak_pw:
                    # weak password Identifier ... coming soon
                    pass
                if perc_pw:
                    self._password_type(word)



            if av_length:
                self.av_wordlen = totalwordlen / self.wordlines
            if perc_pw:
                self._password_type_statistics()
            if dupl:
                output_file.unlink()



    def _password_min_max_info(self, word):
        word_length =len(word)
        if self.smallestword is None or self.biggestword is None:
            self.smallestword = word_length
            self.biggestword = word_length

        if word_length < self.smallestword:
            self.smallestword = word_length
        if word_length > self.biggestword:
            self.biggestword = word_length


    def _char_frequency(self, word):
        for char in word:
            if char in self.char_count:
                self.char_count[char] += 1
            else:
                self.char_count[char] = 1

    def _password_duplicate(self, last_word, duplicate_word):
        if last_word == duplicate_word:
            if duplicate_word not in self.dublicate_words:
                self.dublicate_words.append(duplicate_word)

    def _password_type(self, word: str):
        digit = False
        upper = False
        specsign = False

        for char in word:
            if char.isdigit():
                digit = True
            if char.isupper():
                upper = True
            if char in SPECIAL_CHARACTERS:
                specsign = True

        if digit and upper and specsign:
            self.pw_type_list[self.DIGITUPPERSPECSIGN] += 1
        elif digit and upper:
            self.pw_type_list[self.DIGITANDUPPER] += 1
        elif digit and specsign:
            self.pw_type_list[self.DIGITANDSPECSIGN] += 1
        elif upper and specsign:
            self.pw_type_list[self.UPPERANDSPECSIGN] += 1
        elif digit:
            self.pw_type_list[self.DIGIT] += 1
        elif upper:
            self.pw_type_list[self.UPPER] += 1
        elif specsign:
            self.pw_type_list[self.SPECSIGN] += 1


    def _password_type_statistics(self):
        for type in self.pw_type_list:
            PERCENT_MULTIPLIER = 100
            percent = (self.pw_type_list[type] / self.wordlines) * PERCENT_MULTIPLIER
            self.pw_type_list[type] = percent






    #GETTER---------------------------------------------------------------------------------

    def get_wordlines(self):
        return self.wordlines

    def get_smallest_word(self):
        return self.smallestword

    def get_biggest_word(self):
        return self.biggestword

    def get_average_word_length(self):
        return self.av_wordlen

    def get_occurring_characters(self):
        return self.char_count

    def get_duplicate_words(self):
        return self.dublicate_words

    def get_average_entropy(self):
        #comeing soon...
        pass

    def get_weak_passwords(self):
        #comeing soon...
        pass

    def get_type_occur_percent(self):
        return self.pw_type_list





from pathlib import Path
from textwrap import wrap
from io import StringIO
import sys
from koyaneframework.analyzer.analyzer_general import GeneralAnalyzer
from koyaneframework.analyzer.analyzer_content import ContentAnalyzer
from koyaneframework.load_animation import LoadingSpinner



def print_all_info_wordlist(filepath: Path):
    print_general_info_wordlist(filepath)
    print_content_info_wordlist(filepath)




# general data about the file itself
def print_general_info_wordlist(filepath: Path):
    general = GeneralAnalyzer(filepath)
    print()
    print("---------- GENERAL WORDLIST INFO ----------")
    print(f"File Name.....{general.name()}")
    print(f"Path.....{general.path()}")
    print(f"Size..... {byte_to_megabyte(general.size())} MB / {general.size()} Byte")
    print(f"Last changes..... {general.last_changes()}")
    print(f"Last access..... {general.last_access()}")
    print(f"Created..... {general.created()}")
    print(f"File extension..... {general.extension()}")
    print(f"Estimated encoding..... {general.encoding()}")
    print(f"File hash..... {general.hash()}")
    print()


def print_content_info_wordlist(filepath: Path):
    load = LoadingSpinner("Loading content")
    load.start()
    content = ContentAnalyzer(filepath)
    content.analyze_file(count=True,min_max=True,av_length=True,char_freq=True,av_entropy=False,dupl=True,weak_pw=False,perc_pw=True)
    load.stop()
    print("---------- WORDLIST CONTENT STATISTICS ----------")
    print(f"Total words..... {content.get_wordlines()}")
    print(f"Smallest word..... {content.get_smallest_word()}")
    print(f"Biggest word..... {content.get_biggest_word()}")
    print(f"Average length..... {round(content.get_average_word_length(), 2)}")
    print(f"chars.....{used_chars_to_string(content.get_occurring_characters())}")
    print(f"duplicates found..... {duplicates_to_string(content.get_duplicate_words())}")
    print()

    print("---------- WORDLIST PERCENT STATISTICS ----------")
    print(f"{type_percentage_to_string(content.get_type_occur_percent())}")
    print()


def byte_to_megabyte(byte):
    mb =  1_048_576
    return  round(byte / mb, 2)


from textwrap import wrap


def used_chars_to_string(chars):
    if not chars:
        return "no characters found"

    sorted_chars = sorted(chars)
    all_chars = ",".join(sorted_chars)

    indent = "chars..... "
    wrapped_lines = wrap(all_chars, width=100 - len(indent))  # zieh Einzugslänge ab
    return ("\n" + " " * len(indent)).join([wrapped_lines[0]] + wrapped_lines[1:])


def duplicates_to_string(dup):
    if not dup:
        return "no duplicates found"

    indent = "dublicates found..... "
    words = [word.strip() for word in sorted(dup)]

    lines = []
    for i in range(0, len(words), 10):
        line = ",".join(words[i:i + 10])
        lines.append(line)

    # Indentation: first line with prefix, all subsequent lines indented accordingly.
    return ("\n" + " " * len(indent)).join([lines[0]] + lines[1:])


def type_percentage_to_string(type_list: dict):
    return "\n".join(f"{k}..... = {v:.2f}%" for k, v in type_list.items())




import threading
import time
import sys


class LoadingSpinner:
    def __init__(self, text="Loading", delay=0.3):
        self.text = text
        self.delay = delay
        self.running = False
        self.thread = None
        self.spinner_chars = "|/-\\"
        self.dot_states = ["", ".", "..", "...","....","....."]

    def start(self):
        self.running = True
        self.thread = threading.Thread(target=self._spinner_task)
        self.thread.daemon = True
        self.thread.start()

    def stop(self):
        self.running = False
        if self.thread:
            self.thread.join()
        sys.stdout.write('\r' + ' ' * 80 + '\r')  # Leere Zeile
        sys.stdout.flush()

    def _spinner_task(self):
        i = 0
        j = 0
        next_time = time.perf_counter()
        while self.running:
            spinner = self.spinner_chars[i % len(self.spinner_chars)]
            dots = self.dot_states[j % len(self.dot_states)]
            output = f"\r{spinner} {self.text}{dots}"
            sys.stdout.write(output)
            sys.stdout.flush()

            i += 1
            if i % len(self.spinner_chars) == 0:
                j += 1

            # wait till next tick
            next_time += self.delay
            sleep_time = max(0, next_time - time.perf_counter())
            time.sleep(sleep_time)


# Test
if __name__ == "__main__":
    spinner = LoadingSpinner("Loading infos", delay=0.3)
    spinner.start()

    # Simulates Task
    time.sleep(5)

    spinner.stop()
    print("Finished!")




from koyaneframework.generator.maskChar import MaskChar


class MaskInterpreter:
    mask: str = None
    mask_segments: list[MaskChar] = []  # Python 3.9+

    def __init__(self, mask: str):
        self.mask = mask.strip()
        self.mask_segments = []

        segment = ""
        inside_mask = False

        for char in self.mask:
            if char == "?":
                if inside_mask:  # add built segment
                    self.mask_segments.append(MaskChar(segment))
                    segment = "?"
                else:
                    segment = "?"
                    inside_mask = True
            else:
                segment += char

        if segment.startswith("?"):  # add last segment
            self.mask_segments.append(MaskChar(segment))



import os
import heapq
from pathlib import Path


LOWER_CASE_CHARACTERS: str = "abcdefghijklmnopqrstuvwxyz"       # ?l
UPPER_CASE_CHARACTERS: str = LOWER_CASE_CHARACTERS.upper()      # ?L
LOWER_CASE_VOWELS: str = "aeiou"        # ?v
UPPER_CASE_VOWELS: str = LOWER_CASE_VOWELS.upper()      # ?V
LOWER_CASE_CONSONANTS: str = "bcdfghjklmnpqrstvwxyz"        # ?c
UPPER_CASE_CONSONANTS: str = LOWER_CASE_CONSONANTS.upper()      # ?C

DIGITS: str = "0123456789"      # ?d
SPECIAL_CHARACTERS_MOST_USED: str = "!@#$%^&*()-_+=?"       # ?f
SPECIAL_CHARACTERS_POINTS: str = ".,:;"     # ?p
SPECIAL_CHARACTERS_BRACELET: str = "()[]{}" # ?b

SPECIAL_CHARACTERS: str = "<>|^°!\"§$%&/()=?´{}[]\\¸`+~*#'-_.:,;@€" #?s




temp_dir = Path(__file__).parent / "tmp"
output_file_rem_empty_lines = temp_dir / "removed_empty_lines.kyftmp"

def external_sort(input_file, output_file, chunk_size=1_000_000):


    temp_files = []
    temp_dir = Path(__file__).parent / "tmp" / "chunks"
    temp_dir.mkdir(parents=True, exist_ok=True)

    # Chunks erzeugen
    with open(input_file, 'r', encoding='utf-8', errors='ignore') as f:
        chunk_index = 0
        while True:
            lines = []
            for _ in range(chunk_size):
                line = f.readline()
                if not line:
                    break
                lines.append(line)  # exakt wie eingelesen

            if not lines:
                break

            lines.sort()
            temp_path = temp_dir / f"chunk_{chunk_index}.kyftmp"
            with open(temp_path, 'w', encoding='utf-8') as tf:
                tf.writelines(lines)  # exakt wie eingelesen schreiben

            temp_files.append(temp_path)
            chunk_index += 1

    # Chunks zusammenführen
    files = [open(path, 'r', encoding='utf-8', errors='ignore') for path in temp_files]
    with open(output_file, 'w', encoding='utf-8') as outf:
        iterators = (f for f in files)  # keine Manipulation
        for line in heapq.merge(*iterators):
            outf.write(line)  # direkt schreiben, keine Veränderung

    for f in files:
        f.close()
    # deletes tmp chunk files
    for path in temp_files:
        path.unlink()







def create_new_wordlist(filepath: Path):
    path = os.path.dirname(filepath) or "."

    if os.path.isdir(path):
        with open(filepath, 'w', encoding="utf-8"):
            pass



def add_new_word_to_wordlist(filepath: Path, word: str):
    with open(filepath, "a", encoding="utf-8") as file:
        file.write(f"{word}\n")


def remove_empty_lines(input_path: Path, output_path: Path):
    with input_path.open('r', encoding='utf-8', errors='ignore') as infile, \
         output_path.open('w', encoding='utf-8') as outfile:
        for line in infile:
            if line.strip() != '':
                outfile.write(line)


from pathlib import Path
import datetime
import chardet
import hashlib

class GeneralAnalyzer:
    file_path: Path = None

    def __init__(self, file_path: Path):
        self.file_path = file_path




    def name(self):
        return self.file_path.name

    def path(self):
        return self.file_path.resolve()

    def size(self):
        return self.file_path.stat().st_size

    def last_changes(self):
        return datetime.datetime.fromtimestamp(self.file_path.stat().st_mtime)

    def last_access(self):
        return datetime.datetime.fromtimestamp(self.file_path.stat().st_atime)

    def created(self):
        return datetime.datetime.fromtimestamp(self.file_path.stat().st_ctime)

    def extension(self):
        return self.file_path.suffix

    def encoding(self, sample_size: int = 100_000) -> str:
        with self.file_path.open("rb") as f:
            sample = f.read(sample_size)  #  100.000 Bytes
        result = chardet.detect(sample)
        return result.get("encoding", "Unknown")

    def hash(self):
        hash_md5 = hashlib.md5()
        with self.file_path.open("rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()





"""
maskChar.py – Parsing individual mask segments to determine allowed characters

This file defines the `MaskChar` class, which parses individual mask segments
(e.g., "?l", "?d", "?v", etc.) and determines which character groups are permitted
for password or wordlist generation.

Each instance of `MaskChar` interprets a mask segment, sets flags, and compiles
a string of allowed characters to be used later for string generation.

Masks can represent lowercase letters, uppercase letters, vowels, consonants,
digits, and various classes of special characters.

Example:
MaskChar("?lV") allows all lowercase letters plus all uppercase vowels.

Author: puppetm4ster
"""
import koyaneframework.utils as msk_c
class MaskChar:

    def __init__(self, mask_character: str):
        self.mask_character: str = mask_character.strip().lstrip('?')

        self.all_upper_case: bool = False
        self.all_lower_case: bool = False
        self.all_vowels_lower_case: bool = False
        self.all_vowels_upper_case: bool = False
        self.all_consonants_lower_case: bool = False
        self.all_consonants_upper_case: bool = False
        self.all_digit: bool = False
        self.all_special: bool = False
        self.special_most_used: bool = False
        self.special_points: bool = False
        self.special_bracelet: bool = False

        self.permitted_characters: str = ""

        self._convert_mask_char()


    def _convert_mask_char(self):
        """
            Converts the mask character(s) into internal flags and a character set.

            This method updates flags like `all_lower_case` or `all_digit`, and fills
            `permitted_characters` based on the given mask (e.g., 'l', 'V', 'd', etc.).
            """
        for char in self.mask_character:
            if char == "l":
                self.all_lower_case = True
                self.permitted_characters += msk_c.LOWER_CASE_CHARACTERS
            elif char == "L":
                self.all_upper_case = True
                self.permitted_characters += msk_c.UPPER_CASE_CHARACTERS
            elif char == "v":
                self.all_vowels_lower_case = True
                self.permitted_characters += msk_c.LOWER_CASE_VOWELS
            elif char == "V":
                self.all_vowels_upper_case = True
                self.permitted_characters += msk_c.UPPER_CASE_VOWELS
            elif char == "c":
                self.all_consonants_lower_case = True
                self.permitted_characters += msk_c.LOWER_CASE_CONSONANTS
            elif char == "C":
                self.all_consonants_upper_case = True
                self.permitted_characters += msk_c.UPPER_CASE_CONSONANTS
            elif char == "d":
                self.all_digit = True
                self.permitted_characters += msk_c.DIGITS
            elif char == "s":
                self.all_special = True
                self.permitted_characters += msk_c.SPECIAL_CHARACTERS
            elif char == "f":
                self.special_most_used = True
                self.permitted_characters += msk_c.SPECIAL_CHARACTERS_MOST_USED
            elif char == "p":
                self.special_points = True
                self.permitted_characters += msk_c.SPECIAL_CHARACTERS_POINTS
            elif char == "b":
                self.special_bracelet = True
                self.permitted_characters += msk_c.SPECIAL_CHARACTERS_BRACELET
            else:
                raise ValueError(f"Unknown character: ? {self.mask_character}")





import os
import sys

def load_words_from_file(path):
    """
    reads words from file line by line
    """
    if not os.path.isfile(path):
        print(f" no existing file to read: {path}", file=sys.stderr)
        sys.exit(1)

    with open(path, 'r', encoding='utf-8') as f:
        return [line.strip() for line in f if line.strip()]

def load_chars_from_input(raw_input: str) -> list[str]:
    """
    Splits a string into individual non-whitespace characters.
    """
    return [char for char in raw_input if not char.isspace()]




import typer
from pathlib import Path
from koyaneframework.output_printer import show_banner
from koyaneframework.analyzer.analyzer import print_all_info_wordlist, print_content_info_wordlist, print_general_info_wordlist
from koyaneframework.generator.wordlist_generator import generate_mask_wordlist, generate_wordlist
from koyaneframework.editor.editor import sort_wordlist
from koyaneframework.word_sources import load_chars_from_input, load_words_from_file

app = typer.Typer()

@app.command(help="generates wordlists from scratch")
def generate(
        min_length: int = typer.Option(
            None,
            "-m",
            "--min-length",
            help="specifies the minimum word length."),
        max_length: int = typer.Option(
            None,
            "-x",
            "--max-length",
            help="specifies the maximum word length."),
        mask: str = typer.Option(
            None,
            "-ms",
            "--mask",
            help=("generate wordlist from mask."
                  "A mask is a string consisting of segments. Each segment begins with a “?” followed by one or more letters that define the character type. Example: “?ld?d?f”."
                  "available letter wildcards are “l” = small letter, “L” = capital letter, “v” = small vowel, “V” = capital vowel, “c” = small consonant, “C” = capital consonant, “d” = number,"
                  " “s” = special character, “f” = most commonly used special character, ‘p’ = dot special character and “b” = bracket special character")),
        char_set: str = typer.Option(
            None,
            "-cs",
            "--char_set",
            help=("generate a wordlist with chars."
                  "Example: -cs abcdef1234 || -cs \'#+:_<>abcd123\'")),
        word_file: Path = typer.Option(
            None,
            "-cf",
            "--char_file",
            exists=True,
            dir_okay=False,
            file_okay=True,
            help=("generate a wordlist with a file."
                  "This is useful for word combinations."
                  "The file must contain one char or word per line")),
        output_file: Path = typer.Argument(
            ...,
            exists=False,
            dir_okay=False,
            file_okay=True,
            help="output file")
):

    if mask:    # simple mask generation
        generate_mask_wordlist(mask, output_file)
    elif mask and min_length:   # maskgeneration with min length
        generate_mask_wordlist(mask, output_file, min_len=min_length)
    elif char_set and min_length and max_length:    #char set
        if min_length > max_length:
            raise typer.BadParameter("min_length cannot be greater than max_length")
        else:
            chars = load_chars_from_input(char_set)
            generate_wordlist(chars, min_length, max_length, output_file)
    elif word_file and min_length and max_length:   # word file
        if min_length > max_length:
            raise typer.BadParameter("min_length cannot be greater than max_length")
        else:
            chars = load_words_from_file(word_file)
            generate_wordlist(chars, min_length, max_length, output_file)

@app.command(help="Edit existing word lists")
def edit(
        sort: bool =typer.Option(
            False,
            "--sort",
            "-s",
            help="sort a wordlist"),
        input_file: Path = typer.Argument(
            ...,
            exists=True,
            dir_okay=False,
            file_okay=True,
            readable=True,
            help="Input file which is to be edited"),
        output_file: Path = typer.Argument(
            None,
            exists=False,
            dir_okay=False,
            file_okay=True,
            help="Output file path - default = output/wl.txt")
):
    if output_file is None:
        output_file = Path(__file__).resolve().parent / "output" / "wl.txt"
    else:
        output_file = output_file.resolve()
    output_file.parent.mkdir(parents=True, exist_ok=True)

    show_banner()
    if sort:

        sort_wordlist(input_file,output_file)


@app.command(help="get detailed properties of a wordlist")
def analyze(
        general: bool =typer.Option(
            False,
            "--general",
            "-g",
            help="Print only general info (Filename, Path, Size...)"),
        content: bool = typer.Option(
            False,
            "--content",
            "-c",
            help="Print only content info (Total words, Smallest word, Biggest word...)"
        ),
        file_path: Path = typer.Argument(
            ...,
            exists=True,
            dir_okay=False,
            file_okay=True,
            readable=True,
            help="Input file which is to be analyzed")
):
    show_banner()
    if content and general or not content and not  general:
        pass
        print_all_info_wordlist(file_path)
    elif general:
         print_general_info_wordlist(file_path)

    elif content:
        print_content_info_wordlist(file_path)





@app.command(help="search online for suitable word lists for a specified application and download them (WPA2 ...)")
def search():
    show_banner()
    print("This function is still in work...")


@app.command(help="configurate this script")
def configurate():
    show_banner()





if __name__ == '__main__':
    app()


import itertools
from pathlib import  Path
from koyaneframework.utils import add_new_word_to_wordlist, create_new_wordlist
from koyaneframework.generator.mask_interpreter import MaskInterpreter
from koyaneframework.load_animation import LoadingSpinner


def generate_wordlist(words, min_len, max_len, outfile: Path):
    load = LoadingSpinner(text="Generating wordlist")
    load.start()
    create_new_wordlist(outfile)
    for i in range(min_len, max_len + 1):
        for combination in itertools.permutations(words, i):
            combined = ''.join(combination)
            add_new_word_to_wordlist(outfile, combined)
    load.stop()

def generate_mask_wordlist(mask_arg: str, outfile: Path, max_len: int = None,min_len: int=1):
    load = LoadingSpinner(text="Generating mask wordlist")
    load.start()

    mask = MaskInterpreter(mask_arg)
    if not max_len:
        max_len = len(mask.mask_segments)

    segments = [segment.permitted_characters for segment in mask.mask_segments]

    create_new_wordlist(outfile)
    for i in range(min_len, max_len + 1):
        for combination in itertools.product(*segments):
            word = ''.join(combination)
            add_new_word_to_wordlist(outfile, word)

    load.stop()


# work in progress
def calculate_mask_storage(mask: MaskInterpreter, min_len: int, max_len: int):
    total_combinations = 0
    for length in range(min_len, max_len + 1):
        combinations = 1
        for seg in mask.mask_segments[:length]:
            combinations *= len(seg.permitted_characters)
        total_combinations += combinations
    return total_combinations
